# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles the base interface"""
from abc import ABC, abstractmethod
from typing import Dict

import attr

from spock.backend.wrappers import Spockspace


class BaseInterface(ABC):
    def __init__(self, tuner_config, tuner_namespace: Spockspace):
        """Base init call that maps a few variables

        *Args*:

            _tuner_config: necessary object to determine the interface and sample correctly from the underlying library
            _tuner_namespace: tuner namespace that has attr classes that maps to an underlying library types

        """

        self._tuner_config = {
            k: v for k, v in attr.asdict(tuner_config).items() if v is not None
        }
        self._tuner_namespace = tuner_namespace

    @abstractmethod
    def sample(self):
        """Calls the underlying library sample to get a single sample/draw from the hyper-parameter
        sets (e.g. ranges, choices)

        *Returns*:

            Spockspace of the current hyper-parameter draw

        """
        pass

    @abstractmethod
    def _construct(self):
        """Constructs the base object needed by the underlying library to construct the correct object that allows
        for hyper-parameter sampling

        *Returns*:

            Any typed object needed for support

        """
        pass

    @staticmethod
    def _gen_attr_classes(tune_dict: Dict):
        for k, v in tune_dict.items():
            attrs_dict = {
                ik: attr.ib(
                    validator=attr.validators.instance_of(type(iv)), type=type(iv)
                )
                for ik, iv in v.items()
            }
            obj = attr.make_class(name=k, attrs=attrs_dict, kw_only=True, frozen=True)
            tune_dict.update({k: obj(**v)})
        return tune_dict

    @staticmethod
    def _to_spockspace(tune_dict: Dict):
        """Converts a dict to a Spockspace

        *Args*:

            tune_dict: current dictionary

        *Returns*:

            Spockspace of dict

        """
        return Spockspace(**tune_dict)

    @staticmethod
    def _get_caster(val):
        """Gets a callable type object from a string type

        *Args*:

            val: current attr val:

        *Returns*:

            type class object

        """
        return __builtins__[val.type]

    @property
    @abstractmethod
    def tuner_status(self):
        """Returns a dictionary of all the necessary underlying tuner internals to report the result"""
        pass

    @property
    @abstractmethod
    def best(self):
        """Returns a Spockspace of the best hyper-parameter config and the associated metric value"""



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles the tuner interface interface"""

from typing import Union

from spock.addons.tune.config import OptunaTunerConfig
from spock.addons.tune.optuna import OptunaInterface
from spock.backend.wrappers import Spockspace


class TunerInterface:
    """Handles the general tuner interface by creating the necessary underlying tuner class and dispatches necessary
    ops to the class instance

    *Attributes*:

        _fixed_namespace: fixed parameter namespace used for combination with a sample draw
        _lib_interface: class instance of the underlying hyper-parameter library

    """

    def __init__(
        self,
        tuner_config: Union[OptunaTunerConfig],
        tuner_namespace: Spockspace,
        fixed_namespace: Spockspace,
    ):
        """Init call to the TunerInterface

        *Args*:

            tuner_config: necessary object to determine the interface and sample correctly from the underlying library
            tuner_namespace: tuner namespace that has attr classes that maps to an underlying library types
            fixed_namespace: namespace of fixed parameters

        """
        self._fixed_namespace = fixed_namespace
        # Todo: add ax type check here
        accept_types = OptunaTunerConfig
        if not isinstance(tuner_config, accept_types):
            raise TypeError(
                f"Passed incorrect tuner_config type of {type(tuner_config)} -- must be of type "
                f"{repr(accept_types)}"
            )
        if isinstance(tuner_config, OptunaTunerConfig):
            self._lib_interface = OptunaInterface(
                tuner_config=tuner_config, tuner_namespace=tuner_namespace
            )
        # # TODO: Add ax class logic
        # elif isinstance(tuner_config, (ax.Experiment, ax.SimpleExperiment)):
        #     pass

    def sample(self):
        """Public interface to underlying library sepcific sample that returns a single sample/draw from the
        hyper-parameter sets (e.g. ranges, choices) and combines them with the fixed parameters into a single Spockspace

        *Returns*:

            Spockspace of drawn sample of hyper-parameters and fixed parameters

        """
        curr_sample = self._lib_interface.sample()
        # Merge w/ fixed parameters
        return Spockspace(**vars(curr_sample), **vars(self._fixed_namespace))

    @property
    def tuner_status(self):
        """Returns a dictionary of all the necessary underlying tuner internals to report the result"""
        return self._lib_interface.tuner_status

    @property
    def best(self):
        """Returns a Spockspace of the best hyper-parameter config and the associated metric value"""
        return self._lib_interface.best



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles prepping and saving the Spock config"""

from abc import abstractmethod
from uuid import uuid4

import attr

from spock.backend.handler import BaseHandler
from spock.utils import add_info


class BaseSaver(BaseHandler):  # pylint: disable=too-few-public-methods
    """Base class for saving configs

    Contains methods to build a correct output payload and then writes to file based on the file
    extension

    *Attributes*:

        _writers: maps file extension to the correct i/o handler
        _s3_config: optional S3Config object to handle s3 access

    """

    def __init__(self, s3_config=None):
        super(BaseSaver, self).__init__(s3_config=s3_config)

    def save(
        self,
        payload,
        path,
        file_name=None,
        create_save_path=False,
        extra_info=True,
        file_extension=".yaml",
        tuner_payload=None,
        fixed_uuid=None,
    ):  # pylint: disable=too-many-arguments
        """Writes Spock config to file

        Cleans and builds an output payload and then correctly writes it to file based on the
        specified file extension

        *Args*:

            payload: current config payload
            path: path to save
            file_name: name of file (will be appended with .spock.cfg.file_extension) -- falls back to uuid if None
            create_save_path: boolean to create the path if non-existent
            extra_info: boolean to write extra info
            file_extension: what type of file to write
            tuner_payload: tuner level payload (unsampled)
            fixed_uuid: fixed uuid to allow for file overwrite

        *Returns*:

            None

        """
        # Check extension
        self._check_extension(file_extension=file_extension)
        # Make the filename -- always append a uuid for unique-ness
        uuid_str = str(uuid4()) if fixed_uuid is None else fixed_uuid
        fname = "" if file_name is None else f"{file_name}."
        name = f"{fname}{uuid_str}.spock.cfg{file_extension}"
        # Fix up values -- parameters
        out_dict = self._clean_up_values(payload, file_extension)
        # Fix up the tuner values if present
        tuner_dict = (
            self._clean_tuner_values(tuner_payload)
            if tuner_payload is not None
            else None
        )
        if tuner_dict is not None:
            out_dict.update(tuner_dict)
        # Get extra info
        extra_dict = add_info() if extra_info else None
        try:
            self._supported_extensions.get(file_extension)().save(
                out_dict=out_dict,
                info_dict=extra_dict,
                path=str(path),
                name=name,
                create_path=create_save_path,
                s3_config=self._s3_config,
            )
        except OSError as e:
            print(f"Unable to write to given path: {path / name}")
            raise e

    @abstractmethod
    def _clean_up_values(self, payload, file_extension):
        """Clean up the config payload so it can be written to file

        *Args*:

            payload: dirty payload
            extra_info: boolean to add extra info
            file_extension: type of file to write

        *Returns*:

            clean_dict: cleaned output payload

        """

    @abstractmethod
    def _clean_tuner_values(self, payload):
        """Cleans up the base tuner payload that is not sampled

        *Args*:

            payload: dirty payload

        *Returns*:

            clean_dict: cleaned output payload

        """

    def _clean_output(self, out_dict):
        """Clean up the dictionary so it can be written to file

        *Args*:

            out_dict: cleaned dictionary
            extra_info: boolean to add extra info

        *Returns*:

            clean_dict: cleaned output payload

        """
        # Convert values
        clean_dict = {}
        for key, val in out_dict.items():
            clean_inner_dict = {}
            if isinstance(val, list):
                for idx, list_val in enumerate(val):
                    tmp_dict = {}
                    for inner_key, inner_val in list_val.items():
                        tmp_dict = self._convert(tmp_dict, inner_val, inner_key)
                    val[idx] = tmp_dict
                clean_inner_dict = val
            else:
                for inner_key, inner_val in val.items():
                    clean_inner_dict = self._convert(
                        clean_inner_dict, inner_val, inner_key
                    )
            clean_dict.update({key: clean_inner_dict})
        return clean_dict

    def _convert(self, clean_inner_dict, inner_val, inner_key):
        # Convert tuples to lists so they get written correctly
        if isinstance(inner_val, tuple):
            clean_inner_dict.update(
                {inner_key: self._recursive_tuple_to_list(inner_val)}
            )
        elif inner_val is not None:
            clean_inner_dict.update({inner_key: inner_val})
        return clean_inner_dict

    def _recursive_tuple_to_list(self, value):
        """Recursively turn tuples into lists

        Recursively looks through tuple(s) and convert to lists

        *Args*:

            value: value to check and set typ if necessary
            typed: type of the generic alias to check against

        *Returns*:

            value: updated value with correct type casts

        """
        # Check for __args__ as it signifies a generic and make sure it's not already been cast as a tuple
        # from a composed payload
        list_v = []
        for v in value:
            if isinstance(v, tuple):
                v = self._recursive_tuple_to_list(v)
                list_v.append(v)
            else:
                list_v.append(v)
        return list_v


class AttrSaver(BaseSaver):
    """Base class for saving configs for the attrs backend

    Contains methods to build a correct output payload and then writes to file based on the file
    extension

    *Attributes*:

        _writers: maps file extension to the correct i/o handler

    """

    def __init__(self, s3_config=None):
        super().__init__(s3_config=s3_config)

    def __call__(self, *args, **kwargs):
        return AttrSaver(*args, **kwargs)

    def _clean_up_values(self, payload, file_extension):
        # Dictionary to recursively write to
        out_dict = {}
        # All of the classes are defined at the top level
        all_spock_cls = set(vars(payload).keys())
        out_dict = self._recursively_handle_clean(
            payload, out_dict, all_cls=all_spock_cls
        )
        # Convert values
        clean_dict = self._clean_output(out_dict)
        return clean_dict

    def _clean_tuner_values(self, payload):
        # Just a double nested dict comprehension to unroll to dicts
        out_dict = {
            k: {ik: vars(iv) for ik, iv in vars(v).items()}
            for k, v in vars(payload).items()
        }
        # Convert values
        clean_dict = self._clean_output(out_dict)
        return clean_dict

    def _recursively_handle_clean(
        self, payload, out_dict, parent_name=None, all_cls=None
    ):
        """Recursively works through spock classes and adds clean data to a dictionary

        Given a payload (Spockspace) work recursively through items that don't have parents to catch all
        parameter definitions while correctly mapping nested class definitions to their base level class thus
        allowing the output markdown to be a valid input file

        *Args*:

            payload: current payload (namespace)
            out_dict: output dictionary
            parent_name: name of the parent spock class if nested
            all_cls: all top level spock class definitions

        *Returns*:

            out_dict: modified dictionary with the cleaned data

        """
        for key, val in vars(payload).items():
            val_name = type(val).__name__
            # This catches basic lists and list of classes
            if isinstance(val, list):
                # Check if each entry is a spock class
                clean_val = []
                repeat_flag = False
                for l_val in val:
                    cls_name = type(l_val).__name__
                    # For those that are a spock class and are repeated (cls_name == key) simply convert to dict
                    if (cls_name in all_cls) and (cls_name == key):
                        clean_val.append(attr.asdict(l_val))
                    # For those whose cls is different than the key just append the cls name
                    elif cls_name in all_cls:
                        # Change the flag as this is a repeated class -- which needs to be compressed into a single
                        # k:v pair
                        repeat_flag = True
                        clean_val.append(cls_name)
                    # Fall back to the passed in values
                    else:
                        clean_val.append(l_val)
                # Handle repeated classes
                if repeat_flag:
                    clean_val = list(set(clean_val))[-1]
                out_dict.update({key: clean_val})
            # If it's a spock class but has a parent then just use the class name to reference the values
            elif (val_name in all_cls) and parent_name is not None:
                out_dict.update({key: val_name})
            # Check if it's a spock class without a parent -- iterate the values and recurse to catch more lists
            elif val_name in all_cls:
                new_dict = self._recursively_handle_clean(
                    val, {}, parent_name=key, all_cls=all_cls
                )
                out_dict.update({key: new_dict})
            # Either base type or no nested values that could be Spock classes
            else:
                out_dict.update({key: val})
        return out_dict



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Utility functions for Spock"""

import ast
import os
import re
import socket
import subprocess
import sys
from enum import EnumMeta
from time import localtime, strftime
from warnings import warn

import attr
import git

minor = sys.version_info.minor
if minor < 7:
    from typing import GenericMeta as _GenericAlias
else:
    from typing import _GenericAlias

from typing import Union


def check_path_s3(path: str) -> bool:
    """Checks the given path to see if it matches the s3:// regex

    *Args*:

        path: a spock config path

    *Returns*:

        boolean of regex match

    """
    # Make a case insensitive s3 regex with single or double forward slash (due to posix stripping)
    s3_regex = re.compile(r"(?i)^s3://?").search(path)
    # If it returns an object then the path is an s3 style reference
    return s3_regex is not None


def _is_spock_instance(__obj: object):
    """Checks if the object is a @spock decorated class

    Private interface that checks to see if the object passed in is registered within the spock module and also
    is a class with attrs attributes (__attrs_attrs__)

    *Args*:

        __obj: class to inspect

    *Returns*:

        bool

    """
    return (__obj.__module__ == "spock.backend.config") and attr.has(__obj)


def make_argument(arg_name, arg_type, parser):
    """Make argparser argument based on type

    Based on the type passed in handle the creation of the argparser argument so that overrides will have the correct
    behavior when set

    *Args*:

        arg_name: name for the argument
        arg_type: type of the argument
        parser: current parser

    *Returns*:

        parser: updated argparser

    """
    # For generic alias we take the input string and use a custom type callable to convert
    if isinstance(arg_type, _GenericAlias):
        parser.add_argument(arg_name, required=False, type=_handle_generic_type_args)
    # For Unions -- python 3.6 can't deal with them correctly -- use the same ast method that generics require
    elif (
        hasattr(arg_type, "__origin__")
        and (arg_type.__origin__ is Union)
        and (minor < 7)
    ):
        parser.add_argument(arg_name, required=False, type=_handle_generic_type_args)
    # For choice enums we need to check a few things first
    elif isinstance(arg_type, EnumMeta):
        type_set = list({type(val.value) for val in arg_type})[0]
        # if this is an enum of a class switch the type to str as this is how it gets matched
        type_set = str if type_set.__name__ == "type" else type_set
        parser.add_argument(arg_name, required=False, type=type_set)
    # For booleans we map to store true
    elif arg_type == bool:
        parser.add_argument(arg_name, required=False, action="store_true")
    # Else we are a simple base type which we can cast to
    else:
        parser.add_argument(arg_name, required=False, type=arg_type)
    return parser


def _handle_generic_type_args(val):
    """Evaluates a string containing a Python literal

    Seeing a list and tuple types will come in as string literal format, use ast to get the actual type

    *Args*:

        val: string literal

    *Returns*:

        the underlying string literal type

    """
    return ast.literal_eval(val)


def add_info():
    """Adds extra information to the output dictionary

    *Args*:


    *Returns*:

        out_dict: output dictionary
    """
    out_dict = {}
    out_dict = add_generic_info(out_dict)
    out_dict = add_repo_info(out_dict)
    return out_dict


def make_blank_git(out_dict):
    """Adds blank git info

    *Args*:

        out_dict: current output dictionary

    *Returns*:

        out_dict: output dictionary with added git info

    """
    for key in ("BRANCH", "COMMIT SHA", "STATUS", "ORIGIN"):
        out_dict.update({f"# Git {key}": "UNKNOWN"})
    return out_dict


def add_repo_info(out_dict):
    """Adds GIT information to the output dictionary

    *Args*:

        out_dict: output dictionary

    *Returns*:

        out_dict: output dictionary
    """
    try:  # pragma: no cover
        # Assume we are working out of a repo
        repo = git.Repo(os.getcwd(), search_parent_directories=True)
        # Check if we are really in a detached head state as later info will fail if we are
        if minor < 7:
            head_result = subprocess.run(
                "git rev-parse --abbrev-ref --symbolic-full-name HEAD",
                stdout=subprocess.PIPE,
                shell=True,
                check=False,
            )
        else:
            head_result = subprocess.run(
                "git rev-parse --abbrev-ref --symbolic-full-name HEAD",
                capture_output=True,
                shell=True,
                check=False,
            )
        if head_result.stdout.decode().rstrip("\n") == "HEAD":
            out_dict = make_blank_git(out_dict)
        else:
            out_dict.update({"# Git Branch": repo.active_branch.name})
            out_dict.update({"# Git Commit": repo.active_branch.commit.hexsha})
            out_dict.update(
                {"# Git Date": repo.active_branch.commit.committed_datetime}
            )
            if (
                len(repo.untracked_files) > 0
                or len(repo.active_branch.commit.diff(None)) > 0
            ):
                git_status = "DIRTY"
            else:
                git_status = "CLEAN"
            out_dict.update({"# Git Status": git_status})
            out_dict.update(
                {"# Git Origin": repo.active_branch.commit.repo.remotes.origin.url}
            )
    except git.InvalidGitRepositoryError:  # pragma: no cover
        # But it's okay if we are not
        out_dict = make_blank_git(out_dict)
    return out_dict


def add_generic_info(out_dict):
    """Adds date, fqdn information to the output dictionary

    *Args*:

        out_dict: output dictionary

    *Returns*:

        out_dict: output dictionary
    """
    out_dict.update({"# Machine FQDN": socket.getfqdn()})
    out_dict.update({"# Python Executable": sys.executable})
    out_dict.update(
        {
            "# Python Version": f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
        }
    )
    out_dict.update({"# Python Script": os.path.realpath(sys.argv[0])})
    out_dict.update({"# Run Date": strftime("%Y-%m-%d", localtime())})
    out_dict.update({"# Run Time": strftime("%H:%M:%S", localtime())})
    # Make a best effort to determine if run in a container
    out_dict.update({"# Run w/ Docker": str(_maybe_docker())})
    # Make a best effort to determine if run in a container via k8s
    out_dict.update({"# Run w/ Kubernetes": str(_maybe_k8s())})

    return out_dict


def _maybe_docker(cgroup_path="/proc/self/cgroup"):
    """Make a best effort to determine if run in a docker container

    *Args*:

        cgroup_path: path to cgroup file

    *Returns*:

        boolean of best effort docker determination

    """
    # A few options seem to be at play here:
    # 1. Check for /.dockerenv -- docker should create this is any container
    bool_env = os.path.exists("/.dockerenv")
    # 2. Check /proc/self/cgroup for "docker"
    # https://stackoverflow.com/a/48710609
    bool_cgroup = os.path.isfile(cgroup_path) and any(
        "docker" in line for line in open(cgroup_path)
    )
    return bool_env or bool_cgroup


def _maybe_k8s(cgroup_path="/proc/self/cgroup"):
    """Make a best effort to determine if run in a container via k8s

    *Args*:

        cgroup_path: path to cgroup file

    *Returns*:

        boolean of best effort k8s determination

    """
    # A few options seem to be at play here:
    # 1. Check for KUBERNETES_SERVICE_HOST -- kublet should add this to every running pod
    bool_env = os.environ.get("KUBERNETES_SERVICE_HOST") is not None
    # 2. Similar to docker check /proc/self/cgroup for "kubepods"
    # https://stackoverflow.com/a/48710609
    bool_cgroup = os.path.isfile(cgroup_path) and any(
        "kubepods" in line for line in open(cgroup_path)
    )
    return bool_env or bool_cgroup


def deep_payload_update(source, updates):
    """Deeply updates a dictionary

    Iterates through a dictionary recursively to update individual values within a possibly nested dictionary
    of dictionaries -- creates a dictionary if empty and trying to recurse

    *Args*:

        source: source dictionary
        updates: updates to the dictionary

    *Returns*:

        source: updated version of the source dictionary

    """

    for k, v in updates.items():
        if isinstance(v, dict) and v:
            source_dict = {} if source.get(k) is None else source.get(k)
            updated_dict = deep_payload_update(source_dict, v)
            if updated_dict:
                source[k] = updated_dict
        else:
            source[k] = v
    return source


def check_payload_overwrite(payload, updates, configs, overwrite=""):
    """Warns when parameters are overwritten across payloads as order will matter

    *Args*:

        payload: current payload
        payload_update: update to add to payload
        configs: config path
        overwrite: name of parent

    *Returns*:

    """
    for k, v in updates.items():
        if isinstance(v, dict) and v:
            overwrite += k + ":"
            current_payload = {} if payload.get(k) is None else payload.get(k)
            check_payload_overwrite(current_payload, v, configs, overwrite=overwrite)
        else:
            if k in payload:
                warn(
                    f"Overriding an already set parameter {overwrite + k} from {configs}\n"
                    f"Be aware that value precedence is set by the order of the config files (last to load)...",
                    SyntaxWarning,
                )



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Creates the spock config interface that wraps attr"""

import sys

import attr

from spock.backend.typed import katra


def _base_attr(cls):
    """Map type hints to katras

    Connector function that maps type hinting style to the defined katra style which uses the more strict
    attr.ib() definition

    *Args*:

        cls: basic class def

    *Returns*:

        cls: slotted attrs class that is frozen and kw only
    """
    # Since we are not using the @attr.s decorator we need to get the parent classes for inheritance
    # We do this by using the mro and grabbing anything that is not the first and last indices in the list and wrapping
    # it into a tuple
    if len(cls.__mro__[1:-1]) > 0:
        bases = tuple(cls.__mro__[1:-1])
    # if there are not parents pass a blank tuple
    else:
        bases = ()
    # Make a blank attrs dict for new attrs
    attrs_dict = {}
    if hasattr(cls, "__annotations__"):
        for k, v in cls.__annotations__.items():
            # If the cls has the attribute then a default was set
            if hasattr(cls, k):
                default = getattr(cls, k)
            else:
                default = None
            attrs_dict.update({k: katra(typed=v, default=default)})
    return bases, attrs_dict


def spock_attr(cls):
    """Map type hints to katras

    Connector function that maps type hinting style to the defined katra style which uses the more strict
    attr.ib() definition

    *Args*:

        cls: basic class def

    *Returns*:

        cls: slotted attrs class that is frozen and kw only
    """
    bases, attrs_dict = _base_attr(cls)
    # Dynamically make an attr class
    obj = attr.make_class(
        name=cls.__name__, bases=bases, attrs=attrs_dict, kw_only=True, frozen=True
    )
    # For each class we dynamically create we need to register it within the system modules for pickle to work
    setattr(sys.modules["spock"].backend.config, obj.__name__, obj)
    # Swap the __doc__ string from cls to obj
    obj.__doc__ = cls.__doc__
    return obj



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles the optuna backend"""

import hashlib
import json
from warnings import warn

import attr
import optuna

from spock.addons.tune.config import OptunaTunerConfig
from spock.addons.tune.interface import BaseInterface


class OptunaInterface(BaseInterface):
    """Specific override to support the optuna backend

    *Attributes*:

        _map_type: dictionary that maps class names and types to fns that create optuna distributions
        _tuner_obj: necessary object to determine the interface and sample correctly from the underlying library
        _tuner_namespace: tuner namespace that has attr classes that maps to an underlying library types
        _param_obj: underlying object that optuna study can sample from (flat dictionary)

    """

    def __init__(self, tuner_config: OptunaTunerConfig, tuner_namespace):
        """OptunaInterface init call that maps variables, creates a map to fnc calls, and constructs the necessary
        underlying objects

        *Args*:

            tuner_config: necessary object to determine the interface and sample correctly from the underlying library
            tuner_namespace: tuner namespace that has attr classes that maps to an underlying library types

        """
        super(OptunaInterface, self).__init__(tuner_config, tuner_namespace)
        self._tuner_obj = optuna.create_study(**self._tuner_config)
        self._trial = None
        self._sample_hash = None
        self._trial_status_hash = None
        # Mapping spock underlying classes to optuna distributions (define-and-run interface)
        self._map_type = {
            "RangeHyperParameter": {
                "int": self._uniform_int_dist,
                "float": self._uniform_float_dist,
            },
            "ChoiceHyperParameter": {
                "int": self._categorical_dist,
                "float": self._categorical_dist,
                "str": self._categorical_dist,
                "bool": self._categorical_dist,
            },
        }
        # Build the correct underlying dictionary object for Optuna
        self._param_obj = self._construct()

    @property
    def tuner_status(self):
        return {"trial": self._trial, "study": self._tuner_obj}

    @property
    def best(self):
        rollup_dict, _ = self._trial_rollup(self._tuner_obj.best_trial)
        return (
            self._to_spockspace(self._gen_attr_classes(rollup_dict)),
            self._tuner_obj.best_value,
        )

    def sample(self):
        self._trial = self._tuner_obj.ask(self._param_obj)
        # Roll this back out into a Spockspace so it can be merged into the fixed parameter Spockspace
        # Also need to un-dot the param names to rebuild the nested structure
        rollup_dict, sample_hash = self._trial_rollup(self._trial)
        self._sample_hash = sample_hash
        return self._to_spockspace(self._gen_attr_classes(rollup_dict))

    @staticmethod
    def _trial_rollup(trial):
        """Rollup the trial into a dictionary that can be converted to a spockspace with the correct names and roots

        *Returns*:

            dictionary of rolled up sampled parameters
            md5 hash of the dictionary contents

        """
        key_set = {k.split(".")[0] for k in trial.params.keys()}
        rollup_dict = {val: {} for val in key_set}
        for k, v in trial.params.items():
            split_names = k.split(".")
            rollup_dict[split_names[0]].update({split_names[1]: v})
        dict_hash = hashlib.md5(
            json.dumps(rollup_dict, sort_keys=True).encode("utf-8")
        ).digest()
        return rollup_dict, dict_hash

    def _construct(self):
        """Constructs the base object needed by the underlying library to construct the correct object that allows
        for hyper-parameter sampling

        *Returns*:

            flat dictionary of all hyper-parameters named with dot notation (class.param_name)

        """
        optuna_dict = {}
        # These will only be nested one level deep given the tuner syntax
        for k, v in vars(self._tuner_namespace).items():
            for ik, iv in vars(v).items():
                param_fn = self._map_type[type(iv).__name__][iv.type]
                optuna_dict.update({f"{k}.{ik}": param_fn(iv)})
        return optuna_dict

    @staticmethod
    def _uniform_float_dist(val):
        """Assemble the optuna.distributions.(Log)UniformDistribution object

        https://optuna.readthedocs.io/en/stable/reference/generated/optuna.distributions.UniformDistribution.html
        https://optuna.readthedocs.io/en/stable/reference/generated/optuna.distributions.LogUniformDistribution.html

        *Args*:

            val: current attr val

        *Returns*:

            optuna.distributions.UniformDistribution or optuna.distributions.LogUniformDistribution

        """
        try:
            low = float(val.bounds[0])
            high = float(val.bounds[1])
        except TypeError:
            print(
                f"Attempted to cast into type: {val.type} but failed -- check the inputs to RangeHyperParameter"
            )
        log_scale = val.log_scale
        return (
            optuna.distributions.LogUniformDistribution(low=low, high=high)
            if log_scale
            else optuna.distributions.UniformDistribution(low=low, high=high)
        )

    @staticmethod
    def _uniform_int_dist(val):
        """Assemble the optuna.distributions.Int(Log)UniformDistribution object

        https://optuna.readthedocs.io/en/stable/reference/generated/optuna.distributions.IntUniformDistribution.html
        https://optuna.readthedocs.io/en/stable/reference/generated/optuna.distributions.IntLogUniformDistribution.html

        *Args*:

            val: current attr val

        *Returns*:

            optuna.distributions.IntUniformDistribution or optuna.distributions.IntLogUniformDistribution

        """
        try:
            low = int(val.bounds[0])
            high = int(val.bounds[1])
        except TypeError:
            print(
                f"Attempted to cast into type: {val.type} but failed -- check the inputs to RangeHyperParameter"
            )
        log_scale = val.log_scale
        return (
            optuna.distributions.IntLogUniformDistribution(low=low, high=high)
            if log_scale
            else optuna.distributions.IntUniformDistribution(low=low, high=high)
        )

    def _categorical_dist(self, val):
        """Assemble the optuna.distributions.CategoricalDistribution object

        https://optuna.readthedocs.io/en/stable/reference/generated/optuna.distributions.CategoricalDistribution.html

        *Args*:

            val: current attr val

        *Returns*:

            optuna.distributions.CategoricalDistribution

        """
        caster = self._get_caster(val)
        # Just attempt to cast in a try except
        try:
            val.choices = [caster(v) for v in val.choices]
        except TypeError:
            print(
                f"Attempted to cast into type: {val.type} but failed -- check the inputs to ChoiceHyperParameter"
            )
        return optuna.distributions.CategoricalDistribution(choices=val.choices)



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Creates the spock config decorator that wraps attrs"""

from spock.backend.config import spock_attr
from spock.utils import _is_spock_instance

# Simplified decorator for attrs
spock = spock_attr

# Public alias for checking if an object is a @spock annotated class
isinstance_spock = _is_spock_instance



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""
Spock is a framework that helps manage complex parameter configurations for Python applications

Please refer to the documentation provided in the README.md
"""
from spock.addons.tune.config import (
    ChoiceHyperParameter,
    OptunaTunerConfig,
    RangeHyperParameter,
    spockTuner,
)

__all__ = [
    "builder",
    "config",
    "spockTuner",
    "RangeHyperParameter",
    "ChoiceHyperParameter",
    "OptunaTunerConfig",
]



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Creates the spock config interface that wraps attr -- tune version for hyper-parameters"""
import sys
from typing import List, Optional, Sequence, Tuple, Union

import attr
import optuna

from spock.backend.config import _base_attr


@attr.s(auto_attribs=True)
class OptunaTunerConfig:
    storage: Optional[Union[str, optuna.storages.BaseStorage]] = None
    sampler: Optional[optuna.samplers.BaseSampler] = None
    pruner: Optional[optuna.pruners.BasePruner] = None
    study_name: Optional[str] = None
    direction: Optional[Union[str, optuna.study.StudyDirection]] = None
    load_if_exists: bool = False
    directions: Optional[Sequence[Union[str, optuna.study.StudyDirection]]] = None


def _spock_tune(cls):
    """Ovverides basic spock_attr decorator with another name

    Using a different name allows spock to easily determine which parameters are normal and which are
    meant to be used in a hyper-parameter tuning backend

    *Args*:

        cls: basic class def

    *Returns*:

        cls: slotted attrs class that is frozen and kw only
    """
    bases, attrs_dict = _base_attr(cls)
    # Dynamically make an attr class
    obj = attr.make_class(
        name=cls.__name__, bases=bases, attrs=attrs_dict, kw_only=True, frozen=True
    )
    # For each class we dynamically create we need to register it within the system modules for pickle to work
    setattr(sys.modules["spock"].addons.tune.config, obj.__name__, obj)
    # Swap the __doc__ string from cls to obj
    obj.__doc__ = cls.__doc__
    return obj


# Make the alias for the decorator
spockTuner = _spock_tune


@attr.s
class RangeHyperParameter:
    """Range based hyper-parameter that is sampled uniformly

    Attributes:
        type: type of the hyper-parameter (note: spock will attempt to autocast into this type)
        bounds: min and max of the hyper-parameter range
        log_scale: log scale the values before sampling

    """

    type = attr.ib(
        type=str,
        validator=[
            attr.validators.instance_of(str),
            attr.validators.in_(["float", "int"]),
        ],
    )
    bounds = attr.ib(
        type=Union[Tuple[float, float], Tuple[int, int]],
        validator=attr.validators.deep_iterable(
            member_validator=attr.validators.instance_of((float, int)),
            iterable_validator=attr.validators.instance_of(tuple),
        ),
    )
    log_scale = attr.ib(type=bool, validator=attr.validators.instance_of(bool))


@attr.s
class ChoiceHyperParameter:
    """Choice based hyper-parameter that is sampled uniformly

    Attributes:
        type: type of the hyper-parameter -- (note: spock will attempt to autocast into this type)
        choices: list of variable length that contains all the possible choices to select from

    """

    type = attr.ib(
        type=str,
        validator=[
            attr.validators.instance_of(str),
            attr.validators.in_(["float", "int", "str", "bool"]),
        ],
    )
    choices = attr.ib(
        type=Union[List[str], List[int], List[float], List[bool]],
        validator=attr.validators.deep_iterable(
            member_validator=attr.validators.instance_of((float, int, bool, str)),
            iterable_validator=attr.validators.instance_of(list),
        ),
    )



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles the definitions of arguments types for Spock (backend: attrs)"""

import sys
from enum import EnumMeta
from functools import partial
from typing import TypeVar, Union

import attr

minor = sys.version_info.minor
if minor < 7:
    from typing import GenericMeta as _GenericAlias
else:
    from typing import _GenericAlias


class SavePath(str):
    """Spock special key for saving the Spock config to file

    Defines a special key use to save the current Spock config to file

    """

    def __new__(cls, x):
        return super().__new__(cls, x)


def _get_name_py_version(typed):
    """Gets the name of the type depending on the python version

    *Args*:

        typed: the type of the parameter

    *Returns*:

        name of the type

    """
    return typed._name if hasattr(typed, "_name") else typed.__name__


def _extract_base_type(typed):
    """Extracts the name of the type from a _GenericAlias

    Assumes that the derived types are only of length 1 as the __args__ are [0] recursed... this is not true for
    tuples

    *Args*:

        typed: the type of the parameter

    *Returns*:

        name of type
    """
    if hasattr(typed, "__args__"):
        name = _get_name_py_version(typed=typed)
        bracket_val = f"{name}[{_extract_base_type(typed.__args__[0])}]"
        return bracket_val
    else:
        bracket_value = typed.__name__
    return bracket_value


def _recursive_generic_validator(typed):
    """Recursively assembles the validators for nested generic types

    Walks through the nested type structure and determines whether to recurse all the way to a base type. Once it
    hits the base type it bubbles up the correct validator that is nested within the upper validator

    *Args*:

        typed: input type

    *Returns*:

        return_type: recursively built deep_iterable validators

    """
    if hasattr(typed, "__args__"):
        # If there are more __args__ then we still need to recurse as it is still a GenericAlias
        # Iterate through since there might be multiple types?
        if len(typed.__args__) > 1:
            return_type = attr.validators.deep_iterable(
                member_validator=_recursive_generic_validator(typed.__args__),
                iterable_validator=attr.validators.instance_of(typed.__origin__),
            )
        else:
            return_type = attr.validators.deep_iterable(
                member_validator=_recursive_generic_validator(typed.__args__[0]),
                iterable_validator=attr.validators.instance_of(typed.__origin__),
            )
        return return_type
    else:
        # If no more __args__ then we are to the base type and need to bubble up the type
        # But we need to check against base types and enums
        if isinstance(typed, EnumMeta):
            base_type, allowed = _check_enum_props(typed)
            return_type = attr.validators.and_(
                attr.validators.instance_of(base_type), attr.validators.in_(allowed)
            )
        else:
            return_type = attr.validators.instance_of(typed)
    return return_type


def _generic_alias_katra(typed, default=None, optional=False):
    """Private interface to create a subscripted generic_alias katra

    A 'katra' is the basic functional unit of `spock`. It defines a parameter using attrs as the backend, type checks
    both simple types and subscripted GenericAlias types (e.g. lists and tuples), handles setting default parameters,
    and deals with parameter optionality

    Handles: List[type] and Tuple[type]

    *Args*:

        typed: the type of the parameter to define
        default: the default value to assign if given
        optional: whether to make the parameter optional or not (thus allowing None)

    *Returns*:

        x: Attribute from attrs

    """
    # base python class from which a GenericAlias is derived
    base_typed = typed.__origin__
    if default is not None:
        x = attr.ib(
            validator=_recursive_generic_validator(typed),
            default=default,
            type=base_typed,
            metadata={"base": _extract_base_type(typed), "type": typed},
        )
        # x = attr.ib(validator=_recursive_generic_iterator(typed), default=default, type=base_typed,
        #             metadata={'base': _extract_base_type(typed)})
    elif optional:
        # if there's no default, but marked as optional, then set the default to None
        x = attr.ib(
            validator=attr.validators.optional(_recursive_generic_validator(typed)),
            type=base_typed,
            default=default,
            metadata={
                "optional": True,
                "base": _extract_base_type(typed),
                "type": typed,
            },
        )
        # x = attr.ib(validator=attr.validators.optional(_recursive_generic_iterator(typed)), type=base_typed,
        #             default=default, metadata={'optional': True, 'base': _extract_base_type(typed)})
    else:
        x = attr.ib(
            validator=_recursive_generic_validator(typed),
            type=base_typed,
            metadata={"base": _extract_base_type(typed), "type": typed},
        )
        # x = attr.ib(validator=_recursive_generic_iterator(typed), type=base_typed,
        #             metadata={'base': _extract_base_type(typed)})
    return x


def _check_enum_props(typed):
    """Handles properties of enums

    Checks if all types of the enum are the same and assembles a list of allowed values

    *Args*:

        typed: the type of parameter (Enum)

    *Returns*:

        base_type: the base type of the Enum
        allowed: List of allowed values of the Enum

    """
    # First check if the types of Enum are the same
    type_set = {type(val.value) for val in typed}
    if len(type_set) > 1:
        raise TypeError(f"Enum cannot be defined with multiple types: {type_set}")
    base_type = list(type_set)[-1]
    allowed = [val.value for val in typed]
    return base_type, allowed


def _enum_katra(typed, default=None, optional=False):
    """Private interface to create a Enum typed katra

    A 'katra' is the basic functional unit of `spock`. It defines a parameter using attrs as the backend, type checks
    both simple types and subscripted GenericAlias types (e.g. lists and tuples), handles setting default parameters,
    and deals with parameter optionality

    *Args*:

        typed: the type of the parameter to define
        default: the default value to assign if given
        optional: whether to make the parameter optional or not (thus allowing None)

    *Returns*:

        x: Attribute from attrs

    """
    # First check if the types of Enum are the same
    base_type, allowed = _check_enum_props(typed)
    if base_type.__name__ == "type":
        x = _enum_class_katra(
            typed=typed, allowed=allowed, default=default, optional=optional
        )
    else:
        x = _enum_base_katra(
            typed=typed,
            base_type=base_type,
            allowed=allowed,
            default=default,
            optional=optional,
        )
    return x


def _enum_base_katra(typed, base_type, allowed, default=None, optional=False):
    """Private interface to create a base Enum typed katra

    Here we handle the base types of enums that allows us to force a type check on the instance

    A 'katra' is the basic functional unit of `spock`. It defines a parameter using attrs as the backend, type checks
    both simple types and subscripted GenericAlias types (e.g. lists and tuples), handles setting default parameters,
    and deals with parameter optionality

    *Args*:
        typed: the type of the parameter to define
        base_type: underlying base type
        allowed: set of allowed values
        default: the default value to assign if given
        optional: whether to make the parameter optional or not (thus allowing None)

    *Returns*:

        x: Attribute from attrs

    """
    if default is not None:
        x = attr.ib(
            validator=[
                attr.validators.instance_of(base_type),
                attr.validators.in_(allowed),
            ],
            default=default,
            type=typed,
            metadata={"base": typed.__name__},
        )
    elif optional:
        x = attr.ib(
            validator=attr.validators.optional(
                [attr.validators.instance_of(base_type), attr.validators.in_(allowed)]
            ),
            default=default,
            type=typed,
            metadata={"base": typed.__name__, "optional": True},
        )
    else:
        x = attr.ib(
            validator=[
                attr.validators.instance_of(base_type),
                attr.validators.in_(allowed),
            ],
            type=typed,
            metadata={"base": typed.__name__},
        )
    return x


def _in_type(instance, attribute, value, options):
    """attrs validator for class type enum

    Checks if the type of the class (e.g. value) is in the specified set of types provided

    *Args*:

        instance: current object instance
        attribute: current attribute instance
        value: current value trying to be set in the attrs instance
        options: list, tuple, or enum of allowed options

    *Returns*:

    """
    if type(value) not in options:
        raise ValueError(f"{attribute.name} must be in {options}")


def _enum_class_katra(typed, allowed, default=None, optional=False):
    """Private interface to create a base Enum typed katra

    Here we handle the class based types of enums. Seeing as these classes are generated dynamically we cannot
    force type checking of a specific instance however the in_ validator will catch an incorrect instance type

    A 'katra' is the basic functional unit of `spock`. It defines a parameter using attrs as the backend, type checks
    both simple types and subscripted GenericAlias types (e.g. lists and tuples), handles setting default parameters,
    and deals with parameter optionality

    *Args*:

        typed: the type of the parameter to define
        allowed: set of allowed values
        default: the default value to assign if given
        optional: whether to make the parameter optional or not (thus allowing None)

    *Returns*:

        x: Attribute from attrs

    """
    if default is not None:
        x = attr.ib(
            validator=[partial(_in_type, options=allowed)],
            default=default,
            type=typed,
            metadata={"base": typed.__name__},
        )
    elif optional:
        x = attr.ib(
            validator=attr.validators.optional([partial(_in_type, options=allowed)]),
            default=default,
            type=typed,
            metadata={"base": typed.__name__, "optional": True},
        )
    else:
        x = attr.ib(
            validator=[partial(_in_type, options=allowed)],
            type=typed,
            metadata={"base": typed.__name__},
        )
    return x


def _type_katra(typed, default=None, optional=False):
    """Private interface to create a simple typed katra

    A 'katra' is the basic functional unit of `spock`. It defines a parameter using attrs as the backend, type checks
    both simple types and subscripted GenericAlias types (e.g. lists and tuples), handles setting default parameters,
    and deals with parameter optionality

    Handles: bool, string, float, int, List, and Tuple

    *Args*:

        typed: the type of the parameter to define
        default: the default value to assign if given
        optional: whether to make the parameter optional or not (thus allowing None)

    *Returns*:

        x: Attribute from attrs

    """
    # Grab the name first based on if it is a base type or GenericAlias
    if isinstance(typed, type):
        name = typed.__name__
    elif isinstance(typed, _GenericAlias):
        name = _get_name_py_version(typed=typed)
    else:
        raise TypeError("Encountered an unexpected type in _type_katra")
    special_key = None
    # Default booleans to false and optional due to the nature of a boolean
    if isinstance(typed, type) and name == "bool":
        optional = True
        if default is not True:
            default = False
    # For the save path type we need to swap the type back to it's base class (str)
    elif isinstance(typed, type) and name == "SavePath":
        optional = True
        special_key = name
        typed = str
    if default is not None:
        # if a default is provided, that takes precedence
        x = attr.ib(
            validator=attr.validators.instance_of(typed),
            default=default,
            type=typed,
            metadata={"base": name, "special_key": special_key},
        )
    elif optional:
        x = attr.ib(
            validator=attr.validators.optional(attr.validators.instance_of(typed)),
            default=default,
            type=typed,
            metadata={"optional": True, "base": name, "special_key": special_key},
        )
    else:
        x = attr.ib(
            validator=attr.validators.instance_of(typed),
            type=typed,
            metadata={"base": name, "special_key": special_key},
        )
    return x


def _handle_optional_typing(typed):
    """Handles when a type hint is Optional

    Handles Optional[type] typing and strips out the base type to pass back to the creation of a katra which needs base
    typing

    *Args*:

        typed: type

    *Returns*:

        typed: type (modified if Optional)
        optional: boolean for katra creation
    """
    # Set optional to false
    optional = False
    # Check if it has __args__ to look for optionality as it is a GenericAlias
    if hasattr(typed, "__args__"):
        # If it is more than one than it is most likely optional but check against NoneType in the tuple to verify
        # Check the length of type __args__
        type_args = typed.__args__
        # Optional[X] has type_args = (X, None) and is equal to Union[X, None]
        if (len(type_args) == 2) and (typed == Union[type_args[0], None]):
            typed = type_args[0]
            # Set the optional flag to true
            optional = True
    return typed, optional


def _check_generic_recursive_single_type(typed):
    """Checks generics for the single types -- mixed types of generics are not allowed

    DEPRECATED -- NOW SUPPORTS MIXED TYPES OF TUPLES

    *Args*:

        typed: type

    *Returns*:

    """
    # Check if it has __args__ to look for optionality as it is a GenericAlias
    # if hasattr(typed, '__args__'):
    #     if len(set(typed.__args__)) > 1:
    #         type_list = [str(val) for val in typed.__args__]
    #         raise TypeError(f"Passing multiple different subscript types to GenericAlias is not supported: {type_list}")
    #     else:
    #         for val in typed.__args__:
    #             _check_generic_recursive_single_type(typed=val)
    pass


def katra(typed, default=None):
    """Public interface to create a katra

    A 'katra' is the basic functional unit of `spock`. It defines a parameter using attrs as the backend, type checks
    both simple types and subscripted GenericAlias types (e.g. lists and tuples), handles setting default parameters,
    and deals with parameter optionality

    *Args*:

        typed: the type of the parameter to define
        default: the default value to assign if given
        optional: whether to make the parameter optional or not (thus allowing None)

    Returns:

        x: Attribute from attrs

    """
    # Handle optionals
    typed, optional = _handle_optional_typing(typed)
    # Check generic types for consistent types
    _check_generic_recursive_single_type(typed)
    # We need to check if the type is a _GenericAlias so that we can handle subscripted general types
    # If it is subscript typed it will not be T which python uses as a generic type name
    if isinstance(typed, _GenericAlias) and (
        not isinstance(typed.__args__[0], TypeVar)
    ):
        x = _generic_alias_katra(typed=typed, default=default, optional=optional)
    elif isinstance(typed, EnumMeta):
        x = _enum_katra(typed=typed, default=default, optional=optional)
    else:
        x = _type_katra(typed=typed, default=default, optional=optional)
    return x



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Spock Setup"""

import setuptools
from pkg_resources import parse_requirements

import versioneer

with open("README.md", "r") as fid:
    long_description = fid.read()

with open("REQUIREMENTS.txt", "r") as fid:
    install_reqs = [str(req) for req in parse_requirements(fid)]

with open("./requirements/S3_REQUIREMENTS.txt", "r") as fid:
    s3_reqs = [str(req) for req in parse_requirements(fid)]

with open("./requirements/TUNE_REQUIREMENTS.txt", "r") as fid:
    tune_reqs = [str(req) for req in parse_requirements(fid)]

setuptools.setup(
    name="spock-config",
    description="Spock is a framework designed to help manage complex parameter configurations for Python applications",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="FMR LLC",
    url="https://github.com/fidelity/spock",
    download_url="https://github.com/fidelity/spock",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    project_urls={
        "Source": "https://github.com/fidelity/spock",
        "Documentation": "https://fidelity.github.io/spock/",
        "Bug Tracker": "https://fidelity.github.io/spock/issues",
    },
    keywords=[
        "configuration",
        "argparse",
        "parameters",
        "machine learning",
        "deep learning",
        "reproducibility",
    ],
    packages=setuptools.find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests"]
    ),
    python_requires=">=3.6",
    install_requires=install_reqs,
    extras_require={"s3": s3_reqs, "tune": tune_reqs},
)
